/*
// author : Sven Goethel - Stapenhorststrasse 35a - 33615 Bielefeld
*/

#include "rtti.h"

enum FType { FDefault=0, FCmp, FCopyCreate, FCreate, FDestroy, FReset };

typedef struct rtti_table
{
	rtti              rtype;
  size_t            sz;              /* just to be sure */
	T_FCmpRtti        FCmpRtti;
	T_FCreateRtti     FCreateRtti;
	T_FCopyCreateRtti FCopyCreateRtti;
	T_FDestroyRtti    FDestroyRtti;
  T_FResetRtti      FResetRtti;
  void             *next_entry;
} rtti_table;

static int        NextFreeRttiType=1;
static int        RttiTableEntrys=0;
static rtti_table * RttiTable=NULL;

rtti * NewRttiType(size_t sz)
{
  rtti_table *rt=NULL;
  rtti_table *i=NULL;
  rtti_table *j=RttiTable;
  int CountedRttiTableEntrys=0;

  rt=(rtti_table *)calloc(1, sizeof(rtti_table));
  if(rt==NULL) INT_ERR(__LINE__);
  rt->rtype.sz=sz;
  rt->sz=sz;
  rt->rtype.type=NextFreeRttiType++;

  while (j!=NULL && CountedRttiTableEntrys < RttiTableEntrys )
  {
    i=j;
    j=(rtti_table *)(j->next_entry);
    CountedRttiTableEntrys++;
  }
  if(j!=NULL)
  {
    ErrorOut("this(must be NULL)=%p", j);
    ErrorOut("prev(must be NULL)=%p", i);
    ErrorOut("Entrys=%d, Counted=%d", RttiTableEntrys, CountedRttiTableEntrys);
    INT_ERR(__LINE__);
  }
  if(i==0)
  {
    if(RttiTableEntrys!=0) 
		{ 
		  ErrorOut(GetRttiErrorStr(RTTI_INCORRECT_TABLE));
		  INT_ERR(__LINE__);
    }
    RttiTable=rt;
  }
  else
  {
    i->next_entry=(void *)rt;
  }
  RttiTableEntrys++;
  return &(rt->rtype);
}

static rtti_table *GetRttiTableEntry(const rtti *rtype)
{
  rtti_table *j=RttiTable;
  while(j!=NULL && j->rtype.type!=rtype->type)
    j=(rtti_table *)(j->next_entry);
  if(j==NULL )
	{
		ErrorOut(GetRttiErrorStr(RTTI_UNKNOWN_TYPE));
	  INT_ERR(__LINE__);
  }
  if(j->rtype.sz!=j->sz)
	{
		ErrorOut(GetRttiErrorStr(RTTI_INCORRECT_TABLE));
	  INT_ERR(__LINE__);
  }
  if(rtype->sz != j->sz )
	{
		ErrorOut(GetRttiErrorStr(RTTI_INCORRECT_STRUCTURE));
	  INT_ERR(__LINE__);
  }
  return j;
}

void RegisterRttiFct(const rtti *rtype, T_FCmpRtti cmp, T_FCreateRtti create,
                     T_FCopyCreateRtti cpcreate, T_FDestroyRtti dest,
                     T_FResetRtti reset
                    )
{
  rtti_table *rt=GetRttiTableEntry(rtype);
  if(rt==NULL) INT_ERR(__LINE__);
  rt->FCmpRtti = cmp;
  rt->FCreateRtti = create;
  rt->FCopyCreateRtti = cpcreate;
  rt->FDestroyRtti = dest;
  rt->FResetRtti = reset;
}

static void *GetRttiFunction(const rtti *rtype, enum FType ft)
{
  rtti_table *rt=GetRttiTableEntry(rtype);
	void * result=NULL;

  if(rt==NULL) INT_ERR(__LINE__);

	switch(ft)
	{
		case FCmp: result=(void *)rt->FCmpRtti;
			break;
		case FCreate: result=(void *)rt->FCreateRtti;
			break;
		case FCopyCreate: result=(void *)rt->FCopyCreateRtti;
			break;
		case FDestroy: result=(void *)rt->FDestroyRtti;
			break;
		case FReset: result=(void *)rt->FResetRtti;
			break;
		default : INT_ERR(__LINE__);
			break;
  }
	return result;
}

rtti_result CheckRtti(const rtti *rtype)
{
  rtti_table *j=RttiTable;
  while(j!=NULL && j->rtype.type!=rtype->type)
    j=(rtti_table *)(j->next_entry);
  if(j==NULL ) return RTTI_UNKNOWN_TYPE;
  if(j->rtype.sz!=j->sz) return RTTI_INCORRECT_TABLE;
  if(rtype->sz != j->sz ) return RTTI_INCORRECT_STRUCTURE; 
  return RTTI_OK;
}

const char *GetRttiErrorStr(rtti_result rr)
{
  static char *str[]={ "ok", "incorrect structure", 
	                     "incorrect table",
	                     "unknown type", "no more memory", "wrong error"
										 };
  static char *res="                                                  ";
	static char *err="RTTI ERROR : ";

  strcpy(res,err);
	if(rr>=0 && rr < RTTI_WRONG_ERR) strcat(res, str[rr]);
	else strcat(res, str[RTTI_WRONG_ERR]);
  #ifdef _C166_
    strcat(res, "\n\r");
  #else
    strcat(res, "\n");
  #endif
	return res;
}

void ResetRtti(void)
{
  rtti_table *i=RttiTable;
  rtti_table *j=NULL;
  void *reset;

  while(i!=NULL)
  {
		reset=GetRttiFunction(C_RTTI(i), FReset);
		if(reset) ((T_FResetRtti)reset)();

    j=(rtti_table *)(i->next_entry);
    free(i);
    RttiTableEntrys--;
    i=j;
  }
  if(RttiTableEntrys!=0) INT_ERR(__LINE__);
  RttiTable=NULL;
  NextFreeRttiType=1;
}

int CmpRtti ( const rtti * a, const rtti *b)
{
	void *cmp=NULL;
	int res=0;
	if(a==NULL) INT_ERR(__LINE__);
	if(b==NULL) INT_ERR(__LINE__);

  if(a->type==b->type)
	{
		cmp=GetRttiFunction(a, FCmp);
		if(cmp) res=((T_FCmpRtti)cmp)(a,b);
		else res= memcmp((void *)a,(void *)b,a->sz)==0;
  }
	return res;
}

rtti *CreateRtti ( const rtti * a )
{
	void *create=NULL;
	rtti *element=NULL;
	if(a==NULL) INT_ERR(__LINE__);

	create=GetRttiFunction(a, FCreate);
	if(create) element=((T_FCreateRtti)create)();
	else
  {
    element= (rtti *)calloc(1, a->sz);
		if(element==NULL) INT_ERR(__LINE__);
  }
  element->type=a->type;
  element->sz=a->sz;
	return element;
}

rtti *CopyCreateRtti ( const rtti * a )
{
	void *create=NULL;
	rtti *element=NULL;
	if(a==NULL) INT_ERR(__LINE__);

	create=GetRttiFunction(a, FCopyCreate);
	if(create) element=((T_FCopyCreateRtti)create)(a);
	else
	{
	  element= (rtti *)calloc(1, a->sz);
		if(element==NULL) INT_ERR(__LINE__);
		memcpy((void *)element, (void *)a, a->sz);
  }
  element->type=a->type;
  element->sz=a->sz;
	return element;
}

rtti *AssignRtti ( rtti * a, const rtti *b )
{
  rtti_result rr;
	if(a==NULL) INT_ERR(__LINE__);
	if(b==NULL) INT_ERR(__LINE__);
	if((rr=CheckRtti(a))!=RTTI_OK)
	{
	  ErrorOut(GetRttiErrorStr(rr));
	  INT_ERR(__LINE__);
  }
	if((rr=CheckRtti(b))!=RTTI_OK)
	{
	  ErrorOut(GetRttiErrorStr(rr));
	  INT_ERR(__LINE__);
  }
	if(a->type!=b->type) INT_ERR(__LINE__);

  if(a!=b)
	{
			DestroyRtti(a);
      a=CopyCreateRtti(b);
	}
	return a;
}

void DestroyRtti ( rtti * a )
{
	void *destroy=NULL;
	if(a==NULL) INT_ERR(__LINE__);

	destroy=GetRttiFunction(a, FDestroy);
	if(destroy) ((T_FDestroyRtti)destroy)(a);
	else free((void *)a);
}
